from setuptools import setup, find_packages
setup()
# setup(
#     name='RL_MGM',
#     version='0.0.2',
#     packages=find_packages(),
#     install_requires=[
#         # Add any dependencies here
#     ],
#     entry_points={
#         'console_scripts': [
#             'rl_mgm=rl_mgm.cli:main',
#         ],
#     },
# )
